from .backend import Backend, LocalBackend
from .sql import SqlBackend
from .zmq import ZmqBackend
from .http import HttpBackend

__all__ = [
  'Backend', 'LocalBackend',
  'SqlBackend', 'ZmqBackend', 'HttpBackend',
  
]


from typing_extensions import TypeVar, Generic, Callable, Awaitable, Any, Protocol, overload, Union
from dataclasses import dataclass
import asyncio
from multiprocessing import Process
import traceback
from pydantic import TypeAdapter
from pipeteer.pipelines import Pipeline, Inputtable, Context
from pipeteer.queues import Queue, Transaction, Routed
from pipeteer.util import param_type, return_type, race

Aw = Awaitable
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
AnyT: type = Any # type: ignore

class Stop(Exception):
  ...

class WorkflowContext(Protocol):
  async def call(self, pipe: Inputtable[A, B, Any], x: A, /) -> B:
    ...
  @overload
  async def all(self, a: Aw[A], b: Aw[B], /) -> tuple[A, B]: ...
  @overload
  async def all(self, a: Aw[A], b: Aw[B], c: Aw[C], /) -> tuple[A, B, C]: ...
  @overload
  async def all(self, a: Aw[A], b: Aw[B], c: Aw[C], d: Aw[D], /) -> tuple[A, B, C, D]: ...
  @overload
  async def all(self, *coros: Aw[A]) -> tuple[A, ...]: ...

@dataclass
class WkfContext(WorkflowContext):
  ctx: Context
  states: list
  key: str
  callback_url: str
  step: int = 0

  async def call(self, pipe: Inputtable[A, B, Context], x: A, /) -> B:
    self.step += 1
    if self.step < len(self.states):
      val = self.states[self.step]
      return TypeAdapter(pipe.Tout).validate_python(val)
    else:
      Qin = pipe.input(self.ctx)
      self.ctx.log(f'Calling {pipe.id}({x}), step={self.step}, key="{self.key}"', level='DEBUG')
      await Qin.push(f'{self.step}_{self.key}', { 'url': self.callback_url, 'value': x })
      raise Stop()
    
  async def all(self, *coros: Awaitable):
    n = len(coros)
    if self.step + n < len(self.states):
      prev = self.step+1
      self.step += n
      return tuple(self.states[prev:prev+n])
    
    elif self.step+1 == len(self.states):
      for coro in coros:
        try:
          await coro
        except Stop:
          ...
    raise Stop()
  
@dataclass
class Workflow(Pipeline[A, B, Context, Process], Generic[A, B]):
  call: Callable[[A, WorkflowContext], Awaitable[B]]

  def states(self, ctx: Context):
    return ctx.backend.list_queue(self.id + '-states', tuple[int, Any])

  def urls(self, ctx: Context):
    return ctx.backend.queue(self.id + '-urls', str)
  
  def input(self, ctx: Context) -> Queue[Routed[A]]:
    return ctx.backend.queue(self.id, Routed[self.Tin])
  
  def results(self, ctx: Context) -> tuple[str, Queue]:
    return ctx.backend.public_queue(self.id + '-results', AnyT)
  
  def observe(self, ctx: Context):
    return {
      'input': self.input(ctx),
      'states': self.states(ctx),
      'urls': self.urls(ctx),
      'results': self.results(ctx),
    }

  def run(self, ctx: Context):
      
    self.results(ctx) # trigger creation
    
    async def loop():
      callback_url, Qresults = self.results(ctx)
      Qin = self.input(ctx)
      Qstates = self.states(ctx)
      Qurls = self.urls(ctx)

      async def run(key: str, states: list):
        wkf_ctx = WkfContext(ctx, states=states, key=key, callback_url=callback_url)
        ctx.log(f'Rerunning: key="{key}", states={states}', level='DEBUG')
        out = await self.call(states[0], wkf_ctx)
        ctx.log(f'Outputting: key="{key}", value={out}', level='DEBUG')
        out_url = await Qurls.read(key)
        Qout = ctx.backend.queue_at(out_url, self.Tout)
        
        async with Transaction(Qout, Qurls, Qstates, autocommit=True):
          await Qout.push(key, out)
          await Qurls.pop(key)
          await Qstates.pop(key)

      async def input_step(key, x):
        value, url = x['value'], x['url']
        ctx.log(f'Input loop: key="{key}", value={value}', level='DEBUG')
        try:
          await run(key, [value])
        except Stop:
          async with Transaction(Qin, Qstates, Qurls, autocommit=True):
            await Qin.pop(key)
            await Qstates.push(key, [(0, value)])
            await Qurls.push(key, url)

      async def results_step(idx_key: str, val):
        i, key = idx_key.split('_', 1)
        i = int(i)
        ctx.log(f'Results loop: key="{key}", value={val}, step={i}', level='DEBUG')
        states = await Qstates.read(key) + [(i, val)]
        states = [v for _, v in sorted(states)]

        try:
          await run(key, states)
          await Qresults.pop(idx_key)
          
        except Stop:
          async with Transaction(Qresults, Qstates, autocommit=True):
            await Qresults.pop(idx_key)
            await Qstates.append(key, (i, val))

        # has = await Qresults.has(idx_key)
        # if has:
        #   ctx.log(f'Results loop: key="{key}", value={val}, step={i}, has more', level='CRITICAL')

      while True:
        try:
          idx, (k, v) = await race([Qin.wait_any(), Qresults.wait_any()])
          try:
            fn = input_step if idx == 0 else results_step
            await fn(k, v)

          except Exception:
            loop = 'Input' if idx == 0 else 'Results'
            ctx.log(f'{loop} loop error:', traceback.format_exc(), level='ERROR')

        except Exception:
          ctx.log('Error waiting for items:', traceback.format_exc(), level='ERROR')

    coro = loop()
    return Process(target=asyncio.run, args=(coro,))
  

def workflow(
  *, id: str | None = None,
):
  def decorator(fn: Callable[[A, WorkflowContext], Awaitable[B]]) -> Workflow[A, B]:
    Tin = param_type(fn)
    if Tin is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its input parameter')

    Tout = return_type(fn)
    if Tout is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its return value')
    
    return Workflow(
      Tin=Tin, Tout=Tout,
      id=id or fn.__name__,
      call=fn, # type: ignore
    ) # type: ignore	
  return decorator


from typing_extensions import TypeVar, Generic, AsyncIterable, Callable
from abc import abstractmethod
from dataclasses import dataclass
import asyncio
from datetime import timedelta
from pipeteer.queues import InexistentItem, Transactional, ops

A = TypeVar('A', covariant=True)
B = TypeVar('B', contravariant=True)
C = TypeVar('C')

class ReadQueue(Transactional, Generic[A]):
  """A read/pop-only view of a `Queue`"""

  @abstractmethod
  async def pop(self, key: str, /):
    """Delete a specific item from the queue
    Throws `ReadError`"""

  async def wait_any(self, *, reserve: timedelta | None = None, poll_interval: timedelta = timedelta(seconds=1)) -> tuple[str, A]:
    """Read any item from the queue, waiting if necessary
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `InfraError`
    """
    while not (res := await self.read_any(reserve=reserve)):
      await asyncio.sleep(poll_interval.total_seconds())
    return res

  async def read_any(self, *, reserve: timedelta | None = None) -> tuple[str, A] | None:
    """Read any item from the queue if not empty
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `InfraError`
    """
    async for key, val in self.items(reserve=reserve, max=1):
      return key, val

  @abstractmethod
  async def read(self, key: str, /, *, reserve: timedelta | None = None) -> A:
    """Read a specific item from the queue
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `ReadError`
    """
  
  async def safe_read(self, key: str, /, *, reserve: timedelta | None = None) -> A | None:
    """Read a specific item from the queue
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `InfraError`
    """
    try:
      return await self.read(key, reserve=reserve)
    except InexistentItem:
      ...
  
  @abstractmethod
  async def items(self, *, reserve: timedelta | None = None, max: int | None = None) -> AsyncIterable[tuple[str, A]]:
    """Iterate over the queue's items
    - `reserve`: reservation reserve for each iterated item. If not acknowledged within this time, items will become visible again
    - `max`: maximum number of items to iterate over (and reserve)
    - Throws `InfraError`
    """
    async for key in self.keys():
      yield key, await self.read(key, reserve=reserve)
  
  async def has(self, key: str, /, *, reserve: timedelta | None = None) -> bool:
    """Check if a specific item is in the queue
    - `reserve`: reservation timeout, after which the item is visible again
    - Throws `InfraError`
    """
    return await self.safe_read(key, reserve=reserve) is not None
  
  @abstractmethod
  async def clear(self):
    """Delete all items
    - Throws `InfraError`"""
    
  
  async def keys(self) -> AsyncIterable[str]:
    async for key, _ in self.items(reserve=None, max=None):
      yield key
  
  async def values(self) -> AsyncIterable[A]:
    async for _, val in self.items(reserve=None, max=None):
      yield val

class WriteQueue(Transactional, Generic[B]):
  """A write-only view of a `Queue`"""

  @abstractmethod
  async def push(self, key: str, value: B, /):
    """Push an item into the queue
    Throws `InfraError`"""

  def premap(self, mapper: Callable[[C], B]) -> 'WriteQueue[C]':
    """Map the key-value pair before pushing"""
    async def f(k: str, v: C):
      return k, mapper(v)
    return ops.premap(self, f)
  
  def premap_k(self, mapper: Callable[[str], str]) -> 'WriteQueue[B]':
    """Map the key-value pair before pushing"""
    async def f(k: str, v: B):
      return mapper(k), v
    return ops.premap(self, f)

  def premap_kv(self, mapper: Callable[[str, C], tuple[str, B]]) -> 'WriteQueue[C]':
    """Map the key-value pair before pushing"""
    async def f(k: str, v: C):
      return mapper(k, v)
    return ops.premap(self, f)
  
class Queue(ReadQueue[C], WriteQueue[C], Generic[C]):
  """A key-value, point-readable queue"""

  @staticmethod
  def of(url: str, type: type[A]) -> 'Queue[A]':
    from pipeteer.queues import parse
    return parse(url, type)

  @classmethod
  def sink(cls):
    return Sink()

class ListQueue(Queue[list[C]], Generic[C]):
  @abstractmethod
  async def append(self, key: str, value: C, /):
    ...


class Sink(Queue[C], Generic[C]):
  async def push(self, key: str, value: C):
    ...

  async def pop(self, key: str):
    ...

  async def read(self, key: str, /, *, reserve: timedelta | None = None) -> C:
    raise InexistentItem(key)
  
  async def items(self, *, reserve: timedelta | None = None, max: int | None = None) -> AsyncIterable[tuple[str, C]]:
    if False:
      yield

  async def clear(self):
    ...


from typing_extensions import TypeVar
from dataclasses import dataclass
import asyncio
from sqlalchemy.ext.asyncio import AsyncEngine
from pipeteer.queues import Queue, ListQueue, SqlQueue, ListSqlQueue
from pipeteer.backend import Backend

A = TypeVar('A')

@dataclass
class SqlBackend(Backend):
  engine: AsyncEngine

  def queue(self, id: str, type: type[A]) -> Queue[A]:
    queue = SqlQueue(type, self.engine, table=id)
    asyncio.create_task(queue.initialize())
    return queue
  
  def list_queue(self, id: str, type: type[A]) -> ListQueue[A]:
    return ListSqlQueue(list[type], self.engine, table=id)


from .errors import InexistentItem, InfraError, QueueError, ReadError
from .spec import ReadQueue, WriteQueue, Queue, ListQueue
from .transactions import Transaction, Transactional
from .sql import SqlQueue, ListSqlQueue
from .http import write_api, read_api, queue_api, WriteClient, ReadClient, QueueClient
from .zeromq import ReadZQueue, WriteMQueue, ZeroMQueue
from .routed import Routed, RoutedQueue
from .conn_str import parse
from . import ops, http, zeromq

__all__ = [
  'InexistentItem', 'InfraError', 'QueueError', 'ReadError',
  'ReadQueue', 'WriteQueue', 'Queue', 'ListQueue',
  'Transaction', 'Transactional',
  'SqlQueue', 'ListSqlQueue',
  'Routed', 'RoutedQueue', 'parse',
  'write_api', 'read_api', 'queue_api', 'WriteClient', 'ReadClient', 'QueueClient',
  'ReadZQueue', 'WriteMQueue', 'ZeroMQueue',
  'ops', 'http', 'zeromq',
]


from typing_extensions import TypeVar, Generic, Callable, Awaitable
from dataclasses import dataclass
import asyncio
from datetime import timedelta
from multiprocessing import Process
from pipeteer.pipelines import Pipeline, Context
from pipeteer.queues import Queue, Transaction, Routed
from pipeteer.util import param_type, return_type, num_params, Func1or2

A = TypeVar('A')
B = TypeVar('B')
Ctx = TypeVar('Ctx', bound=Context)

@dataclass
class Activity(Pipeline[A, B, Ctx, Process], Generic[A, B, Ctx]):
  call: Callable[[A, Ctx], Awaitable[B]]
  reserve: timedelta | None = None

  def input(self, ctx: Ctx) -> Queue[Routed[A]]:
    return ctx.backend.queue(self.id, Routed[self.Tin])

  def observe(self, ctx: Ctx) -> dict[str, Queue[Routed[A]]]:
    return { 'input': self.input(ctx) }

  def run(self, ctx: Ctx) -> Process:
    async def loop():
      Qin = self.input(ctx)
      while True:
        try:
          k, x = await Qin.wait_any(reserve=self.reserve)
          ctx.log(f'Processing "{k}" with input {x["value"]}', level='DEBUG')
          try:
            y = await self.call(x['value'], ctx)
            Qout = ctx.backend.queue_at(x['url'], self.Tout)
            async with Transaction(Qin, Qout, autocommit=True):
              await Qout.push(k, y)
              await Qin.pop(k)

          except Exception as e:
            ctx.log(f'Error processing "{k}": {e}. Value: {x["value"]}', level='ERROR')

        except Exception as e:
          ctx.log(f'Error reading from input queue: {e}', level='ERROR')
      
    def runner():
      asyncio.run(loop())

    return Process(target=runner)

def activity(
  id: str | None = None, *,
  reserve: timedelta | None = timedelta(minutes=2),
):
  def decorator(fn: Func1or2[A, Ctx, Awaitable[B]]) -> Activity[A, B, Ctx]:
    Tin = param_type(fn)
    if Tin is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its input parameter')

    Tout = return_type(fn)
    if Tout is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its return value')

    return Activity(
      Tin=Tin or param_type(fn), Tout=Tout or return_type(fn), reserve=reserve, id=id or fn.__name__,
      call=fn if num_params(fn) == 2 else (lambda x, _: fn(x)) # type: ignore
    )
      
  return decorator


from .pipeline import Pipeline, Runnable, Inputtable, Observable, Context
from ._activity import activity, Activity
from ._task import task, Task
from ._multitask import multitask, MultiTask
from ._workflow import workflow, Workflow, WorkflowContext
from ._client import Client
# from ._microservice import microservice, Microservice, Client, ClientContext, Routed

__all__ = [
  'Pipeline', 'Context', 'Runnable', 'Observable', 'Inputtable',
  'activity', 'Activity',
  'task', 'Task',
  'multitask', 'MultiTask',
  'workflow', 'Workflow', 'WorkflowContext',
  'Client',
  # 'microservice', 'Microservice', 'Client',
  # 'ClientContext', 'Routed',
]


from typing_extensions import TypeVar
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pipeteer.queues import Queue, ListQueue

A = TypeVar('A')
B = TypeVar('B')


class Backend(ABC):
  """Backend to create queues"""

  @abstractmethod
  def queue(self, id: str, type: type[A], /) -> Queue[A]:
    ...

  @abstractmethod
  def public_queue(self, id: str, type: type[A], /) -> tuple[str, Queue[A]]:
    ...

  @abstractmethod
  def list_queue(self, id: str, type: type[A], /) -> ListQueue[A]:
    ...

  @abstractmethod
  def queue_at(self, url: str, type: type[A], /) -> Queue[A]:
    ...

  @staticmethod
  def local_sql(url: str):
    from sqlalchemy.ext.asyncio import create_async_engine
    from pipeteer.backend import ZmqBackend
    @dataclass
    class DefaultSqlBackend(LocalBackend, ZmqBackend):
      ...
    return DefaultSqlBackend(id=url, engine=create_async_engine(url))
  
  @staticmethod
  def local_sqlite(path: str):
    return Backend.local_sql(f'sqlite+aiosqlite:///{path}')
  
  @staticmethod
  def sql(*, url: str, sql_url: str):
    from sqlalchemy.ext.asyncio import create_async_engine
    from pipeteer.backend import ZmqBackend, HttpBackend
    @dataclass
    class DefaultSqlBackend(HttpBackend, ZmqBackend):
      ...
    return DefaultSqlBackend(engine=create_async_engine(sql_url), base_url=url)
  
  @staticmethod
  def sqlite(*, url: str, path: str):
    return Backend.sql(url=url, sql_url=f'sqlite+aiosqlite:///{path}')

@dataclass
class LocalBackend(Backend):
  id: str

  def public_queue(self, id: str, type: type[A]) -> tuple[str, Queue[A]]:
    url = f'{self.id}/{id}'
    return url, self.queue(id, type)
  
  def queue_at(self, url: str, type: type[A]) -> Queue[A]:
    if url.startswith(self.id):
      id = url.removeprefix(f'{self.id}/')
      return self.queue(id, type)
    else:
      return Queue.of(url, type)


from pipeteer.queues import Routed, WriteQueue
from typing_extensions import TypeVar, Generic, Any
from dataclasses import dataclass
from pipeteer import Inputtable, Context

A = TypeVar('A')
B = TypeVar('B')

@dataclass
class Client(Inputtable[A, B, Context], Generic[A, B]):
  url: str | None = None

  def input(self, ctx: Context) -> WriteQueue[Routed[A]]:
    if self.url is None:
      raise RuntimeError('Client URL is not set')
    return ctx.backend.queue_at(self.url, Routed[self.Tin])


from .write import tee, prefilter, premap
from .list import singleton, appender
from .delegate import ReadDelegate, WriteDelegate, Delegate

__all__ = [
  'tee', 'prefilter', 'premap',
  'singleton', 'appender',
  'ReadDelegate', 'WriteDelegate', 'Delegate'
]


from datetime import timedelta
import os
from typing_extensions import TypeVar, Generic
from dataclasses import dataclass
from pydantic import BaseModel
import zmq
from zmq.asyncio import Context
from pipeteer.queues import ReadQueue, WriteQueue, Queue, ops
from pipeteer.util import race, exp_backoff

T = TypeVar('T')

class Message(BaseModel, Generic[T]):
  key: str
  value: T

@dataclass
class SubZMQ(Generic[T]):
  topic: str
  type: type[T]
  url: str = 'tcp://localhost:5556'

  def __post_init__(self):
    self.pid = os.getpid()
    ctx = Context.instance()
    self.sub = ctx.socket(zmq.SUB)
    self.sub.connect(self.url)
    self.sub.setsockopt_string(zmq.SUBSCRIBE, self.topic)
  
  async def wait(self) -> tuple[str, T]:
    if os.getpid() != self.pid:
      raise RuntimeError('ReadZQueue is not fork-safe')
    while True:
      msg = await self.sub.recv_string()
      topic, ser_msg = msg.split(' ', 1)
      if topic == self.topic: # could be a prefix but not equal
        msg = Message[self.type].model_validate_json(ser_msg)
        return msg.key, msg.value
  

@dataclass
class ReadZQueue(ops.ReadDelegate[T], Generic[T]):
  """ReadQueue with ZeroMQ realtime notifications"""
  queue: ReadQueue[T]
  sub: SubZMQ[T]

  @classmethod
  def of(cls, queue: ReadQueue[T], topic: str, type: type[T], *, url: str = 'tcp://localhost:5556'):
    return cls(queue, SubZMQ(topic, type, url))

  async def wait_any(self, *, reserve: timedelta | None = None, poll_interval: timedelta = timedelta(seconds=1)):
    if (pair := await self.queue.read_any(reserve=reserve)):
      return pair
    while True:
      _, (key, val) = await race([
        self.sub.wait(),
        exp_backoff(self.queue.read_any, t0=poll_interval.total_seconds(), base=2)
      ])
      if await self.queue.has(key, reserve=reserve):
        return key, val
      


@dataclass
class PubZMQ(Generic[T]):
  topic: str
  type: type[T]
  url: str = 'tcp://localhost:5555'

  @classmethod
  def of(cls, topic: str, type: 'type[T]', *, url: str = 'tcp://localhost:5555'):
    return cls(topic, type, url)
  
  def __post_init__(self):
    self.pid = os.getpid()
    ctx = Context.instance()
    self.pub = ctx.socket(zmq.PUB)
    self.pub.connect(self.url)
  
  async def send(self, key: str, value: T):
    if os.getpid() != self.pid:
      raise RuntimeError('PubZMQ is not fork-safe')
    msg = Message[self.type](key=key, value=value).model_dump_json()
    # print(f'Sending {self.topic}: {msg} to {self.url}')
    self.pub.send_string(f'{self.topic} {msg}')


@dataclass
class WriteMQueue(ops.WriteDelegate[T], Generic[T]):
  """WriteQueue with ZeroMQ realtime notifications"""
  queue: WriteQueue[T]
  pub: PubZMQ[T]

  @classmethod
  def of(cls, queue: WriteQueue[T], topic: str, type: type[T], *, url: str = 'tcp://localhost:5555'):
    return cls(queue, PubZMQ(topic, type, url))

  async def push(self, key: str, value: T):
    await self.queue.push(key, value)
    await self.pub.send(key, value)


@dataclass
class ZeroMQueue(Queue[T], ReadZQueue[T], WriteMQueue[T], Generic[T]):
  """Queue with ZeroMQ realtime notifications"""
  queue: Queue[T] # type: ignore

  @classmethod
  def of( # type: ignore
    cls, queue: Queue[T], topic: str, type: type[T], *,
    pub_url: str = 'tcp://localhost:5555', sub_url: str = 'tcp://localhost:5556'
  ):
    return cls(queue, PubZMQ(topic, type, pub_url), SubZMQ(topic, type, sub_url))


from typing_extensions import TypeVar
from pydantic import TypeAdapter, ValidationError
from fastapi import FastAPI, Request, Response, Body
from fastapi.exceptions import RequestValidationError
from pipeteer.queues import ReadQueue, WriteQueue, Queue, QueueError, InexistentItem

T = TypeVar('T')

def write_api(queue: WriteQueue[T], type: type[T]) -> FastAPI:
  app = FastAPI(generate_unique_id_function=lambda route: route.name)

  parse = TypeAdapter(type).validate_json

  @app.post('/{key:path}', responses={500: {'model': QueueError}, 404: {'model': InexistentItem}})
  async def push(key: str, req: Request, res: Response):
    try:
      value = parse(await req.body())
      await queue.push(key, value)
    except ValidationError as e:
      raise RequestValidationError(e.errors())
    except InexistentItem as e:
      res.status_code = 404
      return e
    except QueueError as e:
      res.status_code = 500
      return e
    
  return app


def read_api(queue: ReadQueue[T]) -> FastAPI:
  app = FastAPI(generate_unique_id_function=lambda route: route.name)

  @app.delete('/item/{key:path}', responses={500: {'model': QueueError}, 404: {'model': InexistentItem}})
  async def pop(key: str, r: Response):
    try:
      return await queue.pop(key)
    except InexistentItem as e:
      r.status_code = 404
      return e
    except QueueError as e:
      r.status_code = 500
      return e
    
  @app.get('/item', responses={500: {'model': QueueError}})
  async def read_any(r: Response) -> tuple[str, T] | None:
    try:
      return await queue.read_any()
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore
    
  @app.get('/item/{key:path}', responses={500: {'model': QueueError}, 404: {'model': InexistentItem}})
  async def read(key: str, r: Response) -> T:
    try:
      return await queue.read(key)
    except InexistentItem as e:
      r.status_code = 404
      return e # type: ignore
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore
    
  @app.get('/keys', responses={500: {'model': QueueError}})
  async def keys(r: Response) -> list[str]:
    try:
      return [key async for key in queue.keys()]
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore
    
  @app.delete('/', responses={500: {'model': QueueError}})
  async def clear(r: Response):
    try:
      await queue.clear()
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore

  return app


def queue_api(queue: Queue[T], type: type[T]) -> FastAPI:
  app = FastAPI(generate_unique_id_function=lambda route: route.name)
  app.mount('/write', write_api(queue, type))
  app.mount('/read', read_api(queue))
  return app


import zmq
from zmq.asyncio import Context#, Socket
import asyncio

async def proxy(
  pub_addr: str = 'tcp://*:5555', sub_addr: str = 'tcp://*:5556',
):
  ctx = Context.instance()

  print(f'Proxying {pub_addr} -> {sub_addr}')

  frontend = ctx.socket(zmq.XSUB)
  frontend.bind(pub_addr)
  
  backend = ctx.socket(zmq.XPUB)
  backend.bind(sub_addr)

  zmq.proxy(frontend, backend)

# async def proxy(
#   pub_addr: str = 'tcp://*:5555', sub_addr: str = 'tcp://*:5556',
# ):
#   ctx = Context.instance()

#   frontend = ctx.socket(zmq.SUB)
#   frontend.bind(pub_addr)
#   frontend.setsockopt_string(zmq.SUBSCRIBE, '')
  
#   backend = ctx.socket(zmq.PUB)
#   backend.bind(sub_addr)

#   print(f'Proxying {pub_addr} -> {sub_addr}')
#   while True:
#     try:
#       msg = await asyncio.wait_for(frontend.recv(), 2)
#       print('Proxying:', msg)
#       await backend.send(msg)
#     except asyncio.TimeoutError:
#       print('Proxy timeout')


from ._zmq import PubZMQ, SubZMQ, ReadZQueue, WriteMQueue, ZeroMQueue, Message
from ._proxy import proxy

__all__ = [
  'PubZMQ', 'SubZMQ', 'ReadZQueue', 'WriteMQueue', 'ZeroMQueue', 'Message', 'proxy'
]


from typing_extensions import TypeVar, Generic, TypedDict, Protocol
from dataclasses import dataclass
from pipeteer.queues import WriteQueue, ReadQueue, Transaction

T = TypeVar('T')
U = TypeVar('U', contravariant=True)

class QueueAt(Protocol, Generic[U]):
  def __call__(self, url: str, /) -> WriteQueue[U]:
    ...

class Routed(TypedDict, Generic[T]):
  url: str
  value: T

@dataclass
class RoutedQueue(WriteQueue[T], Generic[T]):
  Qurls: ReadQueue[str]
  queue_at: QueueAt

  async def push(self, key: str, value: T):
    url = await self.Qurls.read(key)
    Qout = self.queue_at(url)
    async with Transaction(self.Qurls, Qout, autocommit=True):
      await Qout.push(key, value)
      await self.Qurls.pop(key)


import asyncio  
from dataclasses import dataclass

class Transactional:

  async def enter(self, other=None):
    """Enter the transaction
    - Throws `InfraError`
    """
  
  async def commit(self, other=None):
    """Commit the transaction
    - Throws `InfraError`
    """

  async def rollback(self, other=None):
    """Rollback the transaction
    - Throws `InfraError`
    """

  async def close(self, other=None):
    """Close the transaction
    - Throws `InfraError`
    """

  async def __aenter__(self):
    await self.enter()
    return self
  
  async def __aexit__(self, *_):
    await self.close()

@dataclass
class Transaction:
  def __init__(self, agent: Transactional, *agents: Transactional, autocommit: bool = False):
    self.agent = agent
    self.agents = list(agents)
    self.autocommit = autocommit

  async def add(self, agent: Transactional):
    await agent.enter(self.agent)
    self.agents.append(agent)

  async def __aenter__(self):
    await self.agent.enter()
    await asyncio.gather(*[agent.enter(self.agent) for agent in self.agents])
    return self
    
  async def commit(self):
    await asyncio.gather(*[agent.commit(self.agent) for agent in self.agents])
    await self.agent.commit()

  async def close(self):
    await asyncio.gather(*[agent.close(self.agent) for agent in self.agents])
    await self.agent.close()

  async def __aexit__(self, exc_type, exc_value, traceback):
    if exc_type is None and self.autocommit:
      await self.commit()
    await self.close()